/*
 * 
 * The items information is stored into Trie structures, where each structure member is stored
 * into an associative key. It's done like this in case other plugins need to store custom
 * keys into the items for their own purposses.
 * 
 * The basic keys needed for an item to work are the following:
 * 
 * "name"         String      Short, identificative name for the Item. Ex: "demoafro".
 * "long-name"    String      Long name, 'nuff said. Ex: "Demoman's Afro".
 * "model"        String      Model filename.
 * "index"        Integer     TF2's equivalent item definition index.
 * "quality"      Integer     The quality used for this item
 * "level"        Integer     Yes, that's right, the level!
 * "slot"         Integer     Slot number, from 0 to 7.
 * "bodygroups"   Bitflags    Determines wItem bodygroups should be activated/deactivated
 *                            when using this Item.
 * 
 */

// *********************************************************************************
// PREPROCESSOR
// *********************************************************************************
#pragma semicolon 1                  // Force strict semicolon mode.

// *********************************************************************************
// INCLUDES
// *********************************************************************************
#include <sourcemod>
#include <sdktools>
#include <tf2>
#include <tf2_stocks>
#include <adt>

// *********************************************************************************
// CONSTANTS
// *********************************************************************************
// ---- Sourcemod constants --------------------------------------------------------
#define PLUGIN_NAME         "[TF2] Wearable Items [Core]"
#define PLUGIN_AUTHOR       "Damizean"
#define PLUGIN_VERSION      "1.0"
#define PLUGIN_CONTACT      "elgigantedeyeso@gmail.com"
#define CVAR_FLAGS          FCVAR_PLUGIN|FCVAR_NOTIFY

// ---- Plugin constants -----------------------------------------------------------
#define MAXSLOTS            8

#define FLAG_HIDE_SCOUT_HAT        (1 << 0)
#define FLAG_HIDE_SCOUT_HEADPHONES (1 << 1)
#define FLAG_HIDE_HEAVY_HANDS      (1 << 2)
#define FLAG_HIDE_ENGINEER_HELMET  (1 << 3)
#define FLAG_SHOW_SNIPER_QUIVER    (1 << 4)
#define FLAG_HIDE_SNIPER_HAT       (1 << 5)
#define FLAG_HIDE_SOLDIER_ROCKET   (1 << 6)
#define FLAG_HIDE_SOLDIER_HELMET   (1 << 7)
#define FLAG_SHOW_SOLDIER_MEDAL    (1 << 8)

// ---- Engine flags ---------------------------------------------------------------
#define EF_BONEMERGE            (1 << 0)
#define EF_BRIGHTLIGHT          (1 << 1)
#define EF_DIMLIGHT             (1 << 2)
#define EF_NOINTERP             (1 << 3)
#define EF_NOSHADOW             (1 << 4)
#define EF_NODRAW               (1 << 5)
#define EF_NORECEIVESHADOW      (1 << 6)
#define EF_BONEMERGE_FASTCULL   (1 << 7)
#define EF_ITEM_BLINK           (1 << 8)
#define EF_PARENT_ANIMATES      (1 << 9)

// ---- Game bodygroups ------------------------------------------------------------
#define BODYGROUP_SCOUT_HAT        (1 << 0)
#define BODYGROUP_SCOUT_HEADPHONES (1 << 1)
#define BODYGROUP_HEAVY_HANDS      (1 << 0)
#define BODYGROUP_ENGINEER_HELMET  (1 << 0)
#define BODYGROUP_SNIPER_QUIVER    (1 << 0)
#define BODYGROUP_SNIPER_HAT       (1 << 1)
#define BODYGROUP_SOLDIER_ROCKET   (1 << 0)
#define BODYGROUP_SOLDIER_HELMET   (1 << 1)
#define BODYGROUP_SOLDIER_MEDAL    (1 << 2)

//#define DEBUG

enum eItemCodes
{
   Success                  = 1,
   Error_InvalidItem        = -1,
   Error_NotInit            = -2,
   Error_AlreadyExists      = -3,
   Error_NotFound           = -4,
   Error_Cancelled          = -5,
   Error_CanNotUse          = -6,
   Error_InvalidClient      = -7,
   Error_InvalidSlot        = -8
};

// *********************************************************************************
// VARIABLES
// *********************************************************************************
// ---- Forward handles ------------------------------------------------------------
new Handle:g_hForwardOnItemCreated;
new Handle:g_hForwardOnItemDeleted;
new Handle:g_hForwardCanPlayerUseItem;
new Handle:g_hForwardOnPlayerEquipItem;
new Handle:g_hForwardOnPlayerEquipItemPre;
new Handle:g_hForwardOnPlayerRemoveItem;
new Handle:g_hForwardOnPlayerRemoveItemPre;
new Handle:g_hForwardOnItemPreload;

// ---- Item handles ----------------------------------------------------------------
new g_iItemCount;
new Handle:g_hItemTrie;         // The trie structure gives a quick access by name reference.
new Handle:g_hItemTable;        // The array structure allows us to iterate through all the items.

// ---- Client variables -----------------------------------------------------------
new Handle:g_hPlayerItems[MAXPLAYERS+1][MAXSLOTS];
new g_iPlayerEntities[MAXPLAYERS+1][MAXSLOTS];
new g_iPlayerBodyGroups[MAXPLAYERS+1];

// ---- Cvars -----------------------------------------------------------------------

// ---- SDK -------------------------------------------------------------------------
new Handle:g_hSdkEquipWearable;
new Handle:g_hSdkRemoveWearable;

// *********************************************************************************
// PLUGIN
// *********************************************************************************
public Plugin:myinfo =
{
    name        = PLUGIN_NAME,
    author      = PLUGIN_AUTHOR,
    description = PLUGIN_NAME,
    version     = PLUGIN_VERSION,
    url         = PLUGIN_CONTACT
};

// *********************************************************************************
// METHODS
// *********************************************************************************

/*
**
**    ______              
**   / ____/___  ________ 
**  / /   / __ \/ ___/ _ \
** / /___/ /_/ / /  /  __/
** \____/\____/_/   \___/
**
**
*/
    
/* OnPluginStart()
**
** When the plugin is loaded.
** -------------------------------------------------------------------------- */
public OnPluginStart()
{
    // Check if the plugin is being run on the proper mod.
    decl String:strModName[32]; GetGameFolderName(strModName, sizeof(strModName));
    if (!StrEqual(strModName, "tf")) SetFailState("This plugin is only for Team Fortress 2.");
   
    // Create library
    RegPluginLibrary("ItemsCore");
    
    // Create plugin cvars
    CreateConVar("tf_items_version", PLUGIN_VERSION, PLUGIN_NAME, FCVAR_PLUGIN|FCVAR_NOTIFY);
   
    // Startup hats tables
    g_hItemTrie = CreateTrie();
    g_hItemTable = CreateArray();
   
    // Hook all the proper events
    HookEvent("post_inventory_application", Event_InventoryApplication,  EventHookMode_Post);
   
    // Startup SDK calls and external interfacing
    SetupSDK();
    SetupForwards();
}

/* AskPluginLoad()
**
** Registers the plugin's fake natives.
** -------------------------------------------------------------------------- */
public APLRes:AskPluginLoad2(Handle:hMySelf, bool:bLate, String:strError[], iMaxErrors)
{
    CreateNative("CreateItem",               Native_CreateItem);
    CreateNative("DeleteItem",               Native_DeleteItem);
    CreateNative("FindItemByName",           Native_FindItemByName);
    CreateNative("CountItems",               Native_CountItems);
    CreateNative("GetItemFromTable",         Native_GetItemFromTable);
    CreateNative("CountItemSlots",           Native_CountItemSlots);
    CreateNative("GetItemAtSlot",            Native_GetItemAtSlot);
    CreateNative("SetItemAtSlot",            Native_SetItemAtSlot);
    CreateNative("QueryPlayerCanUseItem",    Native_QueryPlayerCanUseItem);
    CreateNative("EquipPlayerWearableItem",  Native_EquipPlayerItem);
    CreateNative("RemovePlayerWearableItem", Native_RemovePlayerItem);
    CreateNative("RemoveWearableItemSlot",   Native_RemovePlayerItemSlot);
    return APLRes_Success;
}

/* SetupSDK()
**
** Registers the SDK methods.
** -------------------------------------------------------------------------- */
stock SetupSDK()
{
    new Handle:hGameConf = LoadGameConfigFile("sm-tf2.items");
    if (hGameConf != INVALID_HANDLE)
    {
        StartPrepSDKCall(SDKCall_Player);
        PrepSDKCall_SetFromConf(hGameConf,SDKConf_Virtual,"EquipWearable");
        PrepSDKCall_AddParameter(SDKType_CBaseEntity, SDKPass_Pointer);
        g_hSdkEquipWearable = EndPrepSDKCall();
        
        StartPrepSDKCall(SDKCall_Player);
        PrepSDKCall_SetFromConf(hGameConf,SDKConf_Virtual,"RemoveWearable");
        PrepSDKCall_AddParameter(SDKType_CBaseEntity, SDKPass_Pointer);
        g_hSdkRemoveWearable = EndPrepSDKCall();
        
        CloseHandle(hGameConf);
    } else {
        SetFailState("Couldn't load SDK functions.");
    }
}

/* SetupForwards()
**
** Registers the plugin's forwards.
** -------------------------------------------------------------------------- */
SetupForwards()
{
    g_hForwardOnItemCreated = CreateGlobalForward("OnItemCreated", ET_Ignore, Param_Cell);
    g_hForwardOnItemDeleted = CreateGlobalForward("OnItemDeleted", ET_Ignore, Param_Cell);
    g_hForwardCanPlayerUseItem = CreateGlobalForward("CanPlayerUseItem", ET_Ignore, Param_Cell, Param_Cell, Param_CellByRef, Param_CellByRef);
    g_hForwardOnPlayerEquipItem = CreateGlobalForward("OnPlayerEquipItem", ET_Ignore, Param_Cell, Param_Cell, Param_Cell);
    g_hForwardOnPlayerEquipItemPre = CreateGlobalForward("OnPlayerEquipItemPre", ET_Hook, Param_Cell, Param_Cell, Param_Cell);
    g_hForwardOnPlayerRemoveItem = CreateGlobalForward("OnPlayerRemoveItem", ET_Ignore, Param_Cell, Param_Cell, Param_Cell);
    g_hForwardOnPlayerRemoveItemPre = CreateGlobalForward("OnPlayerRemoveItemPre", ET_Hook, Param_Cell, Param_Cell, Param_Cell);
    g_hForwardOnItemPreload = CreateGlobalForward("OnItemPreload", ET_Ignore);
}

/* OnMapStart()
**
** When the map is loaded.
** -------------------------------------------------------------------------- */
public OnMapStart()
{
    Forward_OnItemPreload();
    for (new iItem = 0; iItem < g_iItemCount; iItem++)
    {
        new Handle:hItem = Handle:GetArrayCell(g_hItemTable, iItem);
        CacheResources(hItem);
    }
    
    for (new iClient = 1; iClient <= MaxClients; iClient++)
    {
        for (new iSlot; iSlot < MAXSLOTS; iSlot++)
        {
            g_hPlayerItems[iClient][iSlot] = INVALID_HANDLE;
            g_iPlayerEntities[iClient][iSlot] = -1;
        }         
    } 
}
  
/* Event_InventoryApplication()
**
** Evenr fired whenever the player spawns or a resupply locker is used, at this
** point we know the plugin needs to re-create and equip the items again.
** -------------------------------------------------------------------------- */
public Event_InventoryApplication(Handle:hEvent, String:strName[], bool:bDontBroadcast)
{
    new iClient = GetClientOfUserId(GetEventInt(hEvent, "userid"));
    if (!IsValidClient(iClient)) return;
    
    g_iPlayerBodyGroups[iClient] = DefaultBodyGroups(iClient);
    
    for (new iSlot = 0; iSlot < MAXSLOTS; iSlot++)
    {
        new Handle:hItem = g_hPlayerItems[iClient][iSlot];
        if (hItem != INVALID_HANDLE)
        {
            EquipPlayerItem(iClient, g_hPlayerItems[iClient][iSlot], true);
        }
    }
    
    SetEntProp(iClient, Prop_Send, "m_nBody", CalculateBodyGroups(iClient), 4);
}
   
/* EquipPlayerItem()
**
** Equips the selected item onto a player, and creates the proper entities if
** needed. The bRegenerating flag is enabled when the method is called by the
** InventoryApplication event. If successful, returns Success. If not, an error
** code.
** -------------------------------------------------------------------------- */
eItemCodes:EquipPlayerItem(iClient, Handle:hItem, bool:bRegenerating=false)
{
    new bool:bCanUse = true, iTopLevel = -1;
    decl Action:aResult;
    decl iIndex, iQuality, iLevel, iSlot, String:strModelPath[PLATFORM_MAX_PATH];
    decl Handle:hPreviousItem;
    decl iEntity;
   
    // Determine if the client is valid
    if (!IsValidClient(iClient)) return Error_InvalidClient;
   
    // Call the pre-forward, in case something wants to stop this item from being
    // equipped.
    aResult = Forward_OnPlayerEquipItemPre(iClient, hItem, bRegenerating);
    if (aResult == Plugin_Stop) return Error_Cancelled;

    // Query all the plugins if this item
    Forward_CanPlayerUseItem(iClient, hItem, bCanUse, iTopLevel);
    if (bCanUse == false)
    {
        // If the client is regenerating, remove from the slot.
        if (bRegenerating == true)
        {
            RemoveItemOnSlot(iClient, GetItemSlot(hItem));
        }
        return Error_CanNotUse;
    }

    // Now we're sure the player can use this hat, proceed on fetching the
    // base values.
    iIndex   = GetItemIndex(hItem);    
    iQuality = GetItemQuality(hItem);
    iLevel   = GetItemLevel(hItem);
    iSlot    = GetItemSlot(hItem);
    GetItemModel(hItem, strModelPath, sizeof(strModelPath));
   
    // Store the item's handle on the proper slot
    hPreviousItem = g_hPlayerItems[iClient][iSlot];
    g_hPlayerItems[iClient][iSlot] = hItem;
   
    // Retrieve item entity
    iEntity = g_iPlayerEntities[iClient][iSlot];
   
    // Check for the entity, if it doesn't exists, spawn it.
    if (IsPlayerAlive(iClient))
    {   
        new bool:bJustHide = StrEqual(strModelPath, "");
        new bool:bIsItemEntity = IsItemEntityClient(iClient, iEntity, true);
        
        if (bJustHide == true)
        {
            if (bIsItemEntity)
            {
                RemoveItemEntity(iClient, iEntity);
                iEntity = -1;
            }
        }
        else
        {
            if (bIsItemEntity == false)
            {   
                iEntity = CreateItemEntity(iClient, iIndex, iLevel, iQuality);
                EquipItemEntity(iClient, iEntity);
                SetEntityModel(iEntity, strModelPath);
            }
            else if (hItem != hPreviousItem)
            {
                SetEntProp(iEntity, Prop_Send, "m_iItemDefinitionIndex", iIndex);
                SetEntProp(iEntity, Prop_Send, "m_iEntityLevel",         iLevel);
                SetEntProp(iEntity, Prop_Send, "m_iEntityQuality",       iQuality);
                SetEntityModel(iEntity, strModelPath);
            }
        }
        
        // Update bodygroups
        if (bRegenerating == false)
        {
            SetEntProp(iClient, Prop_Send, "m_nBody", CalculateBodyGroups(iClient), 4);
        }
        
        g_iPlayerEntities[iClient][iSlot] = iEntity;
    }
   
    // Done. Publish what we equipped the player item and return.
    Forward_OnPlayerEquipItem(iClient, hItem, bRegenerating);
    return Success;
}

/* RemoveItemOnSlot()
**
** Removes the item equipped on the given slot. The bDeleted flag is activated
** if this method was called after an item destruction (ie. the item was freed
** by another plugin). If successful, returns Success. If not, an error code.
** -------------------------------------------------------------------------- */
eItemCodes:RemoveItemOnSlot(iClient, iSlot, bool:bDeleted=false)
{  
    decl Handle:hItem, Action:aResult;
    
    // Determine if the client is valid, or if the slots are
    if (!IsValidClient(iClient)) return Error_InvalidClient;
    if (iSlot < 0)               return Error_InvalidSlot;
    if (iSlot >= MAXSLOTS)       return Error_InvalidSlot;
    
    // Retrieve the handle to the item
    hItem = g_hPlayerItems[iClient][iSlot];
    if (hItem == INVALID_HANDLE) return Error_NotFound;
	
    // Call the pre-forward, in case something wants to stop this item from being
    // removed.
    if (bDeleted == false)
    {
        aResult = Forward_OnPlayerRemoveItemPre(iClient, hItem, bDeleted);
        if (aResult == Plugin_Stop) return Error_Cancelled;
    }

    // Clear both the entities and the client's slot handle.
    RemoveItemEntity(iClient, g_iPlayerEntities[iClient][iSlot]);
    g_hPlayerItems[iClient][iSlot] = INVALID_HANDLE;
    g_iPlayerEntities[iClient][iSlot] = -1;
    
    // Update body groups
    SetEntProp(iClient, Prop_Send, "m_nBody", CalculateBodyGroups(iClient), 4);
    
    // Done. Publish we removed the player item and return.
    Forward_OnPlayerRemoveItem(iClient, hItem, bDeleted);
    return Success;
}

/*
**
**     ____      __            ____           _            
**    /  _/___  / /____  _____/ __/___ ______(_)___  ____ _
**    / // __ \/ __/ _ \/ ___/ /_/ __ `/ ___/ / __ \/ __ `/
**  _/ // / / / /_/  __/ /  / __/ /_/ / /__/ / / / / /_/ / 
** /___/_/ /_/\__/\___/_/  /_/  \__,_/\___/_/_/ /_/\__, /  
**                                                /____/   
**
**
*/

/* Native_CreateItem()
**
** Fake native method wich obtains an item structure and registers it onto the
** plugin tables.
** -------------------------------------------------------------------------- */
public Native_CreateItem(Handle:hPlugin, iNumParams)
{
    decl Handle:hItem, String:strShortName[64], String:strDummy[2];
    
    // Check if either the item is invalid, uninit, or already exists on the table.
    if ((hItem = Handle:GetNativeCell(1)) == INVALID_HANDLE)                  return _:Error_InvalidItem;    
    if (!GetTrieString(hItem, "name", strShortName, sizeof(strShortName)))    return _:Error_NotInit;
    if (GetTrieString(g_hItemTrie, strShortName, strDummy, sizeof(strDummy))) return _:Error_AlreadyExists;
    
    // Push to reference tables.
    SetTrieValue(g_hItemTrie, strShortName, hItem, true);
    PushArrayCell(g_hItemTable, hItem);
    g_iItemCount++;
    
    // Debug information
    #if defined DEBUG
        decl String:strDebugBuffer[256];
        LogMessage("CreateItem(%x)", hItem);
        GetItemName(hItem,     strDebugBuffer, sizeof(strDebugBuffer)); LogMessage("+-> Short name : %s", strDebugBuffer);
        GetItemLongName(hItem, strDebugBuffer, sizeof(strDebugBuffer)); LogMessage("+-> Long name  : %s", strDebugBuffer);
        GetItemModel(hItem,    strDebugBuffer, sizeof(strDebugBuffer)); LogMessage("+-> Model      : %s", strDebugBuffer);
        LogMessage("+-> Index      : %i", GetItemIndex(hItem));
        LogMessage("+-> Quality    : %i", GetItemQuality(hItem));
        LogMessage("+-> Level      : %i", GetItemLevel(hItem));
        LogMessage("+-> Slot       : %i", GetItemSlot(hItem));
        LogMessage("+-> BodyGroups : %i", GetItemBodyGroups(hItem));
    #endif
    
    // Call forward to announce the creation of the new item.
    Forward_OnItemCreated(hItem);
    return _:Success;
}

/* Native_DeleteItem()
**
** Deletes the given item from the tables and removes it from any player that 
** had it equipped.
** -------------------------------------------------------------------------- */
public Native_DeleteItem(Handle:hPlugin, iNumParams)
{
    decl Handle:hItem, iIndex, String:strShortName[64];

    // Check if either the item is invalid or it's not registered on the tables.
    if ((hItem = Handle:GetNativeCell(1)) == INVALID_HANDLE)               return _:Error_InvalidItem;
    if ((iIndex = FindValueInArray(g_hItemTable, hItem)) == -1)            return _:Error_NotFound;
    if (!GetTrieString(hItem, "name", strShortName, sizeof(strShortName))) return _:Error_NotInit;
    
    // Iterate through all the clients and remove if the item is equipped by
    // any of them.
    decl iSlot;
    GetTrieValue(hItem, "slot", iSlot);
    
    for (new iClient = 1; iClient <= MaxClients; iClient++)
    {
        if (g_hPlayerItems[iClient][iSlot] == hItem)
        {
            RemoveItemOnSlot(iClient, iSlot, true);
        }
    }

    // Destroy from tables
    RemoveFromTrie(g_hItemTrie, strShortName);
    RemoveFromArray(g_hItemTable, iIndex);
    g_iItemCount--;
    
    // Debug information
    #if defined DEBUG
        decl String:strDebugBuffer[256];
        GetItemName(hItem, strDebugBuffer, sizeof(strDebugBuffer));
        LogMessage("DeleteItem(%x) { %s }", hItem, strDebugBuffer);
    #endif
    
    // Call forward to announce the destruction of the new item.
    Forward_OnItemDeleted(hItem);
    CloseHandle(hItem);
    return _:Success;
}

/* Native_FindItemByName()
**
** Looksup the tables to find the given item through it's short name.
** -------------------------------------------------------------------------- */
public Native_FindItemByName(Handle:hPlugin, iNumParams)
{
    decl String:strShortName[32], Handle:hItem;
    
    GetNativeString(1, strShortName, sizeof(strShortName));
    if (GetTrieValue(g_hItemTrie, strShortName, hItem)) 
    {
        // Debug information
        #if defined DEBUG
            LogMessage("FindItemByName(%s) { Found at %x }", strShortName, hItem);
        #endif
        return _:hItem;
    }

    #if defined DEBUG
        LogMessage("FindItemByName(%s) { Not found }", strShortName);
    #endif
    return _:INVALID_HANDLE;
}

/* Native_CountItems()
**
** Returns the item count.
** -------------------------------------------------------------------------- */
public Native_CountItems(Handle:hPlugin, iNumParams)
{
    return g_iItemCount;
}

/* Native_CountItemSlots()
**
** Retrieves the number of slots available for the items.
** -------------------------------------------------------------------------- */
public Native_CountItemSlots(Handle:hPlugin, iNumParams)
{
    return MAXSLOTS;
}

/* Native_GetItemFromTable()
**
** Navigates and returns an item from the table.
** -------------------------------------------------------------------------- */
public Native_GetItemFromTable(Handle:hPlugin, iNumParams)
{
    new iIndex = GetNativeCell(1);
    if (iIndex < 0)                return _:INVALID_HANDLE;
    if (iIndex >= g_iItemCount)    return _:INVALID_HANDLE;
    return _:GetArrayCell(g_hItemTable, iIndex);
}

/* Native_QueryPlayerCanUseItem()
**
** Queries the other plugins to check if the given item can be used by the
** client.
** -------------------------------------------------------------------------- */
public Native_QueryPlayerCanUseItem(Handle:hPlugin, iNumParams)
{
    new iClient = GetNativeCell(1);
    new Handle:hItem = Handle:GetNativeCell(2);
    new bool:bCanUse = true;
    new iTopLevel = 0;
    
    Forward_CanPlayerUseItem(iClient, hItem, bCanUse, iTopLevel);
    return _:bCanUse;
}

/* Native_EquipPlayerItem()
**
** Equips the given item on the player, querying the other plugins for it.
** -------------------------------------------------------------------------- */
public Native_EquipPlayerItem(Handle:hPlugin, iNumParams)
{
    new iClient = GetNativeCell(1);
    new Handle:hItem = Handle:GetNativeCell(2);
    return _:EquipPlayerItem(iClient, hItem);
}

/* Native_RemovePlayerItem()
**
** Removes the given item on the player (if it's equipped)
** -------------------------------------------------------------------------- */
public Native_RemovePlayerItem(Handle:hPlugin, iNumParams)
{
    new iClient = GetNativeCell(1);
    new Handle:hItem = Handle:GetNativeCell(2);    
    new iSlot = -1;
    GetTrieValue(hItem, "slot", iSlot);    
    return _:RemoveItemOnSlot(iClient, iSlot);
}

/* Native_RemovePlayerItemSlot()
**
** Removes the given item on the player (if it's equipped)
** -------------------------------------------------------------------------- */
public Native_RemovePlayerItemSlot(Handle:hPlugin, iNumParams)
{
    new iClient = GetNativeCell(1);
    new iSlot   = GetNativeCell(2);    
    return _:RemoveItemOnSlot(iClient, iSlot);
}

/* Native_SetItemAtSlot()
**
** Sets the equipped item at the given slot.
** -------------------------------------------------------------------------- */
public Native_SetItemAtSlot(Handle:hPlugin, iNumParams)
{
    new iClient      = GetNativeCell(1);
    new iSlot        = GetNativeCell(2);
    new Handle:hItem = GetNativeCell(3);
    g_hPlayerItems[iClient][iSlot] = hItem;
}

/* Native_GetItemAtSlot()
**
** Retrieves the equipped item at the given slot.
** -------------------------------------------------------------------------- */
public Native_GetItemAtSlot(Handle:hPlugin, iNumParams)
{
    new iClient = GetNativeCell(1);
    new iSlot   = GetNativeCell(2);    
    return _:g_hPlayerItems[iClient][iSlot];
}

/* Forward_OnItemCreated()
**
** Called whenever an item has been created and registered in the plugin.
** -------------------------------------------------------------------------- */
Forward_OnItemCreated(Handle:hItem)
{
    Call_StartForward(g_hForwardOnItemCreated);
    Call_PushCell(hItem);
    Call_Finish();
}
 
/* Forward_OnItemDeleted()
**
** Called whenever an item has been deleted.
** -------------------------------------------------------------------------- */
Forward_OnItemDeleted(Handle:hItem)
{
    Call_StartForward(g_hForwardOnItemDeleted);
    Call_PushCell(hItem);
    Call_Finish();
}

/* Forward_CanPlayerUseItem()
**
** Determines if the given item can be used by the client. The reference va-
** riable "bCanUse" tells the plugin if it's permitted to use it or not. The
** "iTopLevel" one is an optional variable in case you want to stablish an
** hierarchy between plugins when deciding if it's usable.
**
** For example, if plugin A with level 4 says the item is usable, but the
** plugin B with level 10 says it isn't, if both plugins respect the hierachical
** levels, the plugin B will be the one to decide if it's useable.
** -------------------------------------------------------------------------- */
Forward_CanPlayerUseItem(iClient, Handle:hItem, &bool:bCanUse, &iTopLevel)
{
    Call_StartForward(g_hForwardCanPlayerUseItem);
    Call_PushCell(iClient);
    Call_PushCell(hItem);
    Call_PushCellRef(bCanUse);
    Call_PushCellRef(iTopLevel);
    Call_Finish();
}

/* Forward_OnPlayerEquipItem()
**
** Called whenever a client successfully equipped an item.
** -------------------------------------------------------------------------- */
Forward_OnPlayerEquipItem(iClient, Handle:hItem, bool:bRegenerating = false)
{
    Call_StartForward(g_hForwardOnPlayerEquipItem);
    Call_PushCell(iClient);
    Call_PushCell(hItem);
    Call_PushCell(bRegenerating);
    Call_Finish();
}

/* Forward_OnPlayerEquipItemPre()
**
** Called whenever a client is about to equip an item. This call is always done
** before the CanPlayerUseItem forward, and returning the action Plugin_Stop will
** inmediately cancel the equip method.
** -------------------------------------------------------------------------- */
Action:Forward_OnPlayerEquipItemPre(iClient, Handle:hItem, bool:bRegenerating = false)
{
    decl Action:aResult;
    Call_StartForward(g_hForwardOnPlayerEquipItemPre);
    Call_PushCell(iClient);
    Call_PushCell(hItem);
    Call_PushCell(bRegenerating);
    Call_Finish(_:aResult);
    return aResult;
}

/* Forward_OnPlayerRemoveItem()
**
** Called whenever a client successfully removed an item.
** -------------------------------------------------------------------------- */
Forward_OnPlayerRemoveItem(iClient, Handle:hItem, bool:bDeleted = false)
{
    Call_StartForward(g_hForwardOnPlayerRemoveItem);
    Call_PushCell(iClient);
    Call_PushCell(hItem);
    Call_PushCell(bDeleted);
    Call_Finish();
}

/* Forward_OnPlayerRemoveItemPre()
**
** Called whenever a client is about to remove an item. Returning the action
** state Plugin_Stop will inmediately cancel the remove method.
** -------------------------------------------------------------------------- */
Action:Forward_OnPlayerRemoveItemPre(iClient, Handle:hItem, bool:bDeleted = false)
{
    decl Action:aResult;
    Call_StartForward(g_hForwardOnPlayerRemoveItemPre);
    Call_PushCell(iClient);
    Call_PushCell(hItem);
    Call_PushCell(bDeleted);
    Call_Finish(_:aResult);
    return aResult;
}

/* Forward_OnItemPreload()
**
** Tells the other plugins we're about to preload.
** -------------------------------------------------------------------------- */
Forward_OnItemPreload()
{
    Call_StartForward(g_hForwardOnItemPreload);
    Call_Finish();
}

/*
** 
**      __  ___            _             __      __  _           
**     /  |/  /___ _____  (_)___  __  __/ /___ _/ /_(_)___  ____ 
**    / /|_/ / __ `/ __ \/ / __ \/ / / / / __ `/ __/ / __ \/ __ \
**   / /  / / /_/ / / / / / /_/ / /_/ / / /_/ / /_/ / /_/ / / / /
**  /_/  /_/\__,_/_/ /_/_/ .___/\__,_/_/\__,_/\__/_/\____/_/ /_/ 
**                      /_/                                       
** 
** 
*/

/* CheckIsValidItem...()
**
** Tests if the given item is valid by checking if the basic tures.
** -------------------------------------------------------------------------- */
stock bool:CheckIsValidItem(Handle:hItem)
{
    decl iTemp, String:strTemp[2];
    if (!GetTrieString(hItem, "name",       strTemp, sizeof(strTemp))) return false;
    if (!GetTrieString(hItem, "long-name",  strTemp, sizeof(strTemp))) return false;
    if (!GetTrieString(hItem, "model",      strTemp, sizeof(strTemp))) return false;
    if (!GetTrieValue(hItem,  "index",      iTemp))                    return false;
    if (!GetTrieValue(hItem,  "quality",    iTemp))                    return false;
    if (!GetTrieValue(hItem,  "level",      iTemp))                    return false;
    if (!GetTrieValue(hItem,  "slot",       iTemp))                    return false;
    if (!GetTrieValue(hItem,  "bodygroups", iTemp))                    return false;
    return true;    
}

/* SetItem...()
**
** Methods to change the basic properties of the item structures.
** -------------------------------------------------------------------------- */
stock SetItemName(Handle:hItem, String:strBuffer[])     { SetTrieString(hItem, "name",       strBuffer);   }
stock SetItemLongName(Handle:hItem, String:strBuffer[]) { SetTrieString(hItem, "long-name",  strBuffer);   }
stock SetItemModel(Handle:hItem, String:strBuffer[])    { SetTrieString(hItem, "model",      strBuffer);   }
stock SetItemIndex(Handle:hItem, iIndex)                { SetTrieValue(hItem,  "index",      iIndex);      }
stock SetItemQuality(Handle:hItem, iQuality)            { SetTrieValue(hItem,  "quality",    iQuality);    }
stock SetItemLevel(Handle:hItem, iLevel)                { SetTrieValue(hItem,  "level",      iLevel);      }
stock SetItemSlot(Handle:hItem, iSlot)                  { SetTrieValue(hItem,  "slot",       iSlot);       }
stock SetItemBodyGroups(Handle:hItem, iBodyGroups)      { SetTrieValue(hItem,  "bodygroups", iBodyGroups); }

/* GetItem...()
**
** Methods to retrieve the basic properties of the item structures.
** -------------------------------------------------------------------------- */
stock GetItemName(Handle:hItem, String:strBuffer[], iMaxLength)     { GetTrieString(hItem, "name",      strBuffer, iMaxLength); }
stock GetItemLongName(Handle:hItem, String:strBuffer[], iMaxLength) { GetTrieString(hItem, "long-name", strBuffer, iMaxLength); }
stock GetItemModel(Handle:hItem, String:strBuffer[], iMaxLength)    { GetTrieString(hItem, "model",     strBuffer, iMaxLength); }
stock GetItemIndex(Handle:hItem)                                    { decl iValue; GetTrieValue(hItem, "index",      iValue); return iValue; }
stock GetItemQuality(Handle:hItem)                                  { decl iValue; GetTrieValue(hItem, "quality",    iValue); return iValue; }
stock GetItemLevel(Handle:hItem)                                    { decl iValue; GetTrieValue(hItem, "level",      iValue); return iValue; }
stock GetItemSlot(Handle:hItem)                                     { decl iValue; GetTrieValue(hItem, "slot",       iValue); return iValue; }
stock GetItemBodyGroups(Handle:hItem)                               { decl iValue; GetTrieValue(hItem, "bodygroups", iValue); return iValue; }

/*
**
**   ______            __    
**  /_  __/___  ____  / /____
**   / / / __ \/ __ \/ / ___/
**  / / / /_/ / /_/ / (__  ) 
** /_/  \____/\____/_/____/  
**
**
*/

/* CreateItemEntity()
**
** Creates a valid wearable item entity.
** -------------------------------------------------------------------------- */
stock CreateItemEntity(iClient, iIndex, iLevel, iQuality)
{
    new iTeam = GetClientTeam(iClient);
    new iItem = CreateEntityByName("tf_wearable_item");
   
    if (IsValidEdict(iItem))
    {
        SetEntProp(iItem, Prop_Send, "m_bInitialized", 1);
        
        SetEntProp(iItem, Prop_Send, "m_fEffects",             EF_BONEMERGE|EF_BONEMERGE_FASTCULL|EF_NOSHADOW|EF_PARENT_ANIMATES);
        SetEntProp(iItem, Prop_Send, "m_iTeamNum",             iTeam);
        SetEntProp(iItem, Prop_Send, "m_nSkin",                (iTeam-2));
        SetEntProp(iItem, Prop_Send, "m_CollisionGroup",       11);
        SetEntProp(iItem, Prop_Send, "m_iItemDefinitionIndex", iIndex);
        SetEntProp(iItem, Prop_Send, "m_iEntityLevel",         iLevel);
        SetEntProp(iItem, Prop_Send, "m_iEntityQuality",       iQuality);
        
        DispatchSpawn(iItem);
    }
    
    return iItem;
}

/* EquipItemEntity()
**
** Equips the given wearable item entity.
** -------------------------------------------------------------------------- */
stock EquipItemEntity(iClient, iItem)
{
    if (IsItemEntity(iItem))
    {
        SDKCall(g_hSdkEquipWearable, iClient, iItem);
    }
}

/* RemoveItemEntity()
**
** Removes the given wearable item entity.
** -------------------------------------------------------------------------- */
stock RemoveItemEntity(iClient, iItem)
{
    if (IsItemEntity(iItem))
    {
        if (GetEntPropEnt(iItem, Prop_Send, "m_hOwnerEntity") == iClient)
        {
            SDKCall(g_hSdkRemoveWearable, iClient, iItem);
        }
        RemoveEdict(iItem);
    }
}

/* IsItemEntity()
**
** Checks if the given entity is a valid wearable item.
** -------------------------------------------------------------------------- */
stock bool:IsItemEntity(iEntity)
{
    if (iEntity > 0)
    {
        if (IsValidEdict(iEntity))
        {
            new String:strClassname[32];
            GetEdictClassname(iEntity, strClassname, sizeof(strClassname));
            return StrEqual(strClassname, "tf_wearable_item", false);
        }
    }
    
    return false;
}

/* IsItemEntityClient()
**
** Checks if the given entity is a valid wearable item. It also checks if it's
** attached to a client.
** -------------------------------------------------------------------------- */
stock bool:IsItemEntityClient(iClient, iEntity, bool:bRemove = false)
{
    if (IsItemEntity(iEntity))
    {
        if (GetEntPropEnt(iEntity, Prop_Send, "m_hOwnerEntity") != iClient)
        {
            if (bRemove) RemoveEdict(iEntity);
            return false;
        }
        return true;
    }
    
    return false;
}

/* IsValidClient()
**
** Determines if the given client is valid (valid index, connected and in-game).
** -------------------------------------------------------------------------- */
bool:IsValidClient(iClient)
{
    if (iClient < 0 || iClient > MaxClients) return false;
    if (!IsClientConnected(iClient)) return false;
    return IsClientInGame(iClient);
}

/* DefaultBodyGroups()
**
** Retrieves the base body group bits depending upon the class and weapons.
** -------------------------------------------------------------------------- */
DefaultBodyGroups(iClient)
{
    new TFClassType:tfcClass = TF2_GetPlayerClass(iClient);
    if (tfcClass == TFClass_Sniper || tfcClass == TFClass_Heavy)
    {
        return (GetEntProp(iClient, Prop_Send, "m_nBody", 4) & 1);
    }
    return 0;
}

/* CalculateBodyGroups()
**
** Determine the final body groups after applying the items.
** -------------------------------------------------------------------------- */
CalculateBodyGroups(iClient)
{
    new iBodyGroups = g_iPlayerBodyGroups[iClient];
    new iItemGroups = 0;
    
    for (new iSlot = 0; iSlot < MAXSLOTS; iSlot++)
    {
        if (g_hPlayerItems[iClient][iSlot] == INVALID_HANDLE) continue;
        iItemGroups |= GetItemBodyGroups(g_hPlayerItems[iClient][iSlot]);
    }
    
    switch(TF2_GetPlayerClass(iClient))
    {
        case TFClass_Heavy:
        {
            if (iItemGroups & FLAG_HIDE_HEAVY_HANDS) iBodyGroups |= BODYGROUP_HEAVY_HANDS;
        }
        case TFClass_Engineer:
        {
            if (iItemGroups & FLAG_HIDE_ENGINEER_HELMET) iBodyGroups |= BODYGROUP_ENGINEER_HELMET;
        }
        case TFClass_Scout:
        {
            if (iItemGroups & FLAG_HIDE_SCOUT_HAT) iBodyGroups |= BODYGROUP_SCOUT_HAT;
            if (iItemGroups & FLAG_HIDE_SCOUT_HEADPHONES) iBodyGroups |= BODYGROUP_SCOUT_HEADPHONES;
        }
        case TFClass_Sniper:
        {
            if (iItemGroups & FLAG_SHOW_SNIPER_QUIVER) iBodyGroups |= BODYGROUP_SNIPER_QUIVER;
            if (iItemGroups & FLAG_HIDE_SNIPER_HAT) iBodyGroups |= BODYGROUP_SNIPER_HAT;
        }
        case TFClass_Soldier:
        {
            if (iItemGroups & FLAG_HIDE_SOLDIER_ROCKET) iBodyGroups |= BODYGROUP_SOLDIER_ROCKET;   
            if (iItemGroups & FLAG_HIDE_SOLDIER_HELMET) iBodyGroups |= BODYGROUP_SOLDIER_HELMET;   
            if (iItemGroups & FLAG_SHOW_SOLDIER_MEDAL) iBodyGroups |= BODYGROUP_SOLDIER_MEDAL;   
        }
    }
    
    return iBodyGroups;
}

/* CacheResources()
**
** Precaches and adds to download table the resources of an item.
** -------------------------------------------------------------------------- */
CacheResources(Handle:hItem)
{
    if (hItem == INVALID_HANDLE) return;

    decl String:strKey[32];
    decl String:strFile[PLATFORM_MAX_PATH];    
    
    for (new iResource;;iResource++)
    {
        // Determine resource key
        Format(strKey, sizeof(strKey), "resource-%i", iResource);
        
        // Retrieve and check validness.
        if (!GetTrieString(hItem, strKey, strFile, sizeof(strFile))) break;
        if (!FileExists(strFile, false) && !FileExists(strFile, true)) continue;
        
        // Precache and upload to table.
        if (StrContains(strFile, ".vmt", false) != -1)      PrecacheDecal(strFile, true);
        else if (StrContains(strFile, ".mdl", false) != -1) PrecacheModel(strFile, true);
        else if (StrContains(strFile, ".pcf", false) != -1) PrecacheGeneric(strFile, true);
        AddFileToDownloadsTable(strFile);
    }
}